# SPDX-License-Identifier: GPL-2.0
# Copyright (C) 2025 Bardia Moshiri <bardia@furilabs.com>

import subprocess
from pathlib import Path
import threading

_command_lock = threading.Lock()
HELPER_PATH = "/usr/libexec/bootman-helper"

def run_helper(*args):
    """
    Run the bootman helper script with the given arguments.

    Args:
        *args: Arguments to pass to the helper script

    Returns:
        tuple: (success_boolean, message)
    """
    global _command_lock

    with _command_lock:  # Ensure only one privileged operation runs at a time
        try:
            result = subprocess.run(
                ['pkexec', HELPER_PATH, *map(str, args)],
                capture_output=True,
                text=True
            )

            if result.returncode != 0:
                return False, f"Command execution failed: {result.stderr}"
            return True, result.stdout
        except Exception as e:
            return False, f"Error executing command: {str(e)}"

def mount_partition():
    """Mount the FuriOS persist partition."""
    return run_helper("mount")

def get_partition_size(partition_name):
    """Get the size of a specific LVM partition."""
    success, output = run_helper("lvdisplay", f"/dev/droidian/{partition_name}")

    if success:
        for line in output.splitlines():
            if "LV Size" in line:
                size = line.split()[2]
                unit = line.split()[3]
                return f"{size}{unit}"
    return "Unknown"

def write_partitions_file(partitions):
    """Write partitions to /furios_persist/bootman/partitions file."""
    content = ''
    for partition in partitions:
        display_name = process_partition_name(partition)
        content += f"{partition}:{display_name}\n"

    return run_helper("write_partitions", content)

def create_install_commands(name, size):
    """Create commands for creating a new partition."""
    partition_name = name.replace(" ", "-").lower()

    # Get current root filesystem size
    success, output = run_helper("lvdisplay", "/dev/droidian/droidian-rootfs")
    if not success:
        return False, "Failed to get current partition size"

    current_size = None
    for line in output.splitlines():
        if "LV Size" in line:
            current_size = float(line.split()[2])
            break

    if current_size is None:
        return False, "Could not determine current partition size"

    # Calculate new sizes
    new_size = current_size - float(size)
    new_size_mb = int(new_size * 1024)
    size_mb = int(float(size) * 1024)

    # Write the command file
    commands = [
        "e2fsck -fy /dev/droidian/droidian-rootfs",
        f"resize2fs /dev/droidian/droidian-rootfs {new_size_mb}M",
        f"lvm lvreduce -L -{size_mb}M -r /dev/droidian/droidian-rootfs",
        f"lvm lvcreate -L {size_mb}M -n {partition_name} droidian",
        f"mke2fs /dev/droidian/{partition_name}",
        f"e2fsck -fy /dev/droidian/{partition_name}"
    ]
    success, _ = run_helper("write_commands", "\n".join(commands))
    if not success:
        return False, "Failed to write commands"

    # Write the wip file
    success, _ = run_helper("write_wip", f"{partition_name}:{name}")
    if not success:
        return False, "Failed to write wip file"

    return True, "Partition creation queued successfully"

def delete_install_commands(partition_name):
    """Create commands for deleting a partition."""
    # Get partition size
    success, output = run_helper("lvdisplay", f"/dev/droidian/{partition_name}")
    if not success:
        return False, "Failed to get partition size"

    size = None
    for line in output.splitlines():
        if "LV Size" in line:
            size = float(line.split()[2])
            unit = line.split()[3]
            if unit.lower() == 'gib':
                size = size * 1024
            break

    if size is None:
        return False, "Could not determine partition size"

    # Remove WIP
    success, output = run_helper("remove_wip")
    if not success:
        return False, "Failed to remove wip"

    # Write the command file
    commands = [
        f"lvm lvremove -f /dev/droidian/{partition_name}",
        f"lvm lvextend -L +{int(size)}M /dev/droidian/droidian-rootfs",
        "resize2fs /dev/droidian/droidian-rootfs"
    ]
    success, _ = run_helper("write_commands", "\n".join(commands))
    if not success:
        return False, "Failed to write commands"

    # Update partitions file
    success, output = run_helper("cat", "/furios_persist/bootman/partitions")
    if success:
        partitions = [line.strip() for line in output.splitlines() if partition_name not in line]
        content = "\n".join(partitions)
        success, _ = run_helper("write_partitions", content)
        if not success:
            return False, "Failed to update partitions file"

    return True, "Deletion queued successfully"

def is_mounted(mount_point):
    """Check if a mount point is currently mounted."""
    try:
        return Path('/proc/mounts').read_text().find(mount_point) != -1
    except Exception:
        return False

def is_partition_mounted(partition_name):
    """Check if a specific partition is currently mounted."""
    try:
        return Path('/proc/mounts').read_text().find(f"/dev/droidian/{partition_name}") != -1
    except Exception:
        return False

def process_partition_name(partition_name):
    """Convert a partition name to a more readable format."""
    name = partition_name.replace('droidian-', '').replace('furios-', '')
    words = name.split('-')
    return ' '.join(word.capitalize() for word in words)

def list_partitions():
    """List all available partitions."""
    try:
        droidian_path = Path("/dev/droidian")
        if not droidian_path.exists():
            return []

        excluded = ['droidian-persistent', 'droidian-reserved']
        return [p.name for p in droidian_path.iterdir()
                if p.exists() and p.name not in excluded]
    except Exception:
        return []

def read_partitions_file(partitions_file):
    """Read and parse the partitions file."""
    try:
        content = Path(partitions_file).read_text().strip()
        return [line.strip().split(':')
                for line in content.split('\n')
                if ':' in line]
    except Exception:
        return []

def get_queued_partition():
    """
    Check for any queued partition operations.

    Returns:
        tuple: (operation, partition_name, display_name) where:
               operation is either 'install' or 'delete'
               partition_name is the name of the partition
               display_name is the friendly name for display
        None: if no queue exists
    """
    try:
        wip_file = Path("/furios_persist/bootman/wip-partitions")
        commands_file = Path("/furios_persist/bootman/commands")

        # Check if commands file exists
        if not commands_file.exists():
            return None

        commands_content = commands_file.read_text().strip()
        if not commands_content:
            return None

        # Check if it's an installation (both files exist)
        if wip_file.exists():
            wip_content = wip_file.read_text().strip()
            if wip_content and ':' in wip_content:
                partition_name, display_name = wip_content.split(':', 1)
                return ('install', partition_name.strip(), display_name.strip())

        # If no wip file but commands exist, check for deletion
        for line in commands_content.split('\n'):
            if 'lvremove' in line:
                # Extract partition name from command like "lvm lvremove -f /dev/droidian/partition-name"
                partition_name = line.split('/')[-1].strip()
                display_name = partition_name.replace('-', ' ')
                return ('delete', partition_name, display_name)

        return None
    except Exception:
        return None

def get_supported_operating_systems():
    """
    Get list of supported operating systems.

    Returns:
        list: List of tuples containing (name, description, icon_name)
    """
    return [
        (
            "FuriOS",
            "FuriOS is a Linux OS for mobile devices from FuriLabs",
            "computer-symbolic"
        )
    ]

def get_os_download_url(os_name):
    """
    Get the download URL for a specific operating system.

    Args:
        os_name (str): Name of the operating system

    Returns:
        str: URL to download the OS image, or None if not found
    """
    url_map = {
        "FuriOS": "https://filedump.furios.io/krypton/vendor-1.img"
    }
    return url_map.get(os_name)
